#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the LatticeDecoder class.
"""

import logging
import math

import numpy
import theano
from theano import tensor

from theanolm.backend import InputError
from theanolm.backend import interpolate_linear, interpolate_loglinear
from theanolm.backend import logprob_type
from theanolm.network import RecurrentState

class LatticeDecoder(object):
    """Word Lattice Decoding Using a Neural Network Language Model
    """

    class Token:
        """Decoding Token

        A token represents a partial path through a word lattice. The decoder
        propagates a set of tokens through the lattice by
        """
        __slots__ = ("history", "state", "ac_logprob", "lat_lm_logprob",
                     "nn_lm_logprob", "recombination_hash", "graph_logprob",
                     "total_logprob")

        def __init__(self,
                     history=(),
                     state=None,
                     ac_logprob=logprob_type(0.0),
                     lat_lm_logprob=logprob_type(0.0),
                     nn_lm_logprob=logprob_type(0.0)):
            """Constructs a token with given recurrent state and logprobs.

            The constructor won't compute the total logprob. The user is
            responsible for computing it when necessary, to avoid unnecessary
            overhead.

            New tokens will not have recombination hash and total log
            probability set.

            :type history: list of ints
            :param history: word IDs that the token has passed

            :type state: RecurrentState
            :param state: the state of the recurrent layers for a single
                          sequence

            :type ac_logprob: logprob_type
            :param ac_logprob: sum of the acoustic log probabilities of the
                               lattice links

            :type lat_lm_logprob: logprob_type
            :param lat_lm_logprob: sum of the LM log probabilities of the
                                   lattice links

            :type nn_lm_logprob: logprob_type
            :param nn_lm_logprob: sum of the NNLM log probabilities of the
                                  lattice links
            """

            self.history = history
            self.state = [] if state is None else state
            self.ac_logprob = ac_logprob
            self.lat_lm_logprob = lat_lm_logprob
            self.nn_lm_logprob = nn_lm_logprob
            self.recombination_hash = None
            self.graph_logprob = None
            self.total_logprob = None

        @classmethod
        def copy(cls, token):
            """Creates a copy of a token.

            The recurrent layer states will not be copied - a pointer will be
            copied instead. There's no need to copy the structure, since we
            never modify the state of a token, but replace it if necessary.

            Recombination hash and total log probability will not be copied.

            :type token: LatticeDecoder.Token
            :param token: a token to copy

            :rtype: LatticeDecoder.Token
            :returns: a copy of ``token``
            """

            return cls(token.history,
                       token.state,
                       token.ac_logprob,
                       token.lat_lm_logprob,
                       token.nn_lm_logprob)

        def recompute_hash(self, recombination_order):
            """Computes the hash that will be used to decide if two tokens
            should be recombined.

            :type recombination_order: int
            :param recombination_order: number of words to consider when
                recombining tokens, or ``None`` for the entire history
            """

            if recombination_order is None:
                limited_history = self.history
            else:
                limited_history = self.history[-recombination_order:]
            self.recombination_hash = hash(limited_history)

        def recompute_total(self, nn_lm_weight, lm_scale, wi_penalty,
                            linear=False):
            """Computes the interpolated language model log probability and
            the total log probability.

            First the interpolated LM log probability is computed. Then the
            total log probability is computed by applying the LM scale factor,
            and adding the acoustic log probability and the word insertion
            penalty.

            :type nn_lm_weight: logprob_type
            :param nn_lm_weight: weight of the neural network LM probability
                                 when interpolating with the lattice probability

            :type lm_scale: logprob_type
            :param lm_scale: scaling factor for LM probability when computing
                             the total probability

            :type wi_penalty: logprob_type
            :param wi_penalty: penalize each word in the history by adding this
                               value as many times as there are words

            :type linear: bool
            :param linear: if set to ``True`` performs linear interpolation
                           instead of (pseudo) log-linear
            """

            if linear:
                lm_logprob = interpolate_linear(
                    self.nn_lm_logprob, self.lat_lm_logprob,
                    nn_lm_weight)
            else:
                lm_logprob = interpolate_loglinear(
                    self.nn_lm_logprob, self.lat_lm_logprob,
                    nn_lm_weight, (1.0 - nn_lm_weight))

            self.graph_logprob = lm_logprob * lm_scale
            self.graph_logprob += wi_penalty * (len(self.history) - 1)

            self.total_logprob = self.ac_logprob + self.graph_logprob

        def history_words(self, vocabulary):
            """Converts the word IDs in the history to words using
            ``vocabulary``. The history may contain also OOV words as text, so
            any ``str`` will be left untouched.

            :type vocabulary: Vocabulary
            :param vocabulary: mapping from word IDs to words

            :rtype: list of strs
            :returns: the token's history as list of words
            """

            return [vocabulary.id_to_word[word] if isinstance(word, int)
                    else word
                    for word in self.history]

        def __str__(self, vocabulary=None):
            """Creates a string representation of the token.

            :type vocabulary: Vocabulary
            :param vocabulary: if a vocabulary is given, uses it to decode
                               history word names from word IDs

            :rtype: str
            :returns: a string that includes all the attributes in one line
            """

            if vocabulary is None:
                history = ' '.join(str(x) for x in self.history)
            else:
                history = ' '.join(self.history_words(vocabulary))

            if self.total_logprob is None:
                return '[{}]  acoustic: {:.2f}  lattice LM: {:.2f}  NNLM: ' \
                       '{:.2f}'.format(
                           history,
                           self.ac_logprob,
                           self.lat_lm_logprob,
                           self.nn_lm_logprob)
            else:
                return '[{}]  acoustic: {:.2f}  lattice LM: {:.2f}  NNLM: ' \
                       '{:.2f}  total: {:.2f}'.format(
                           history,
                           self.ac_logprob,
                           self.lat_lm_logprob,
                           self.nn_lm_logprob,
                           self.total_logprob)

    def __init__(self, network, decoding_options, profile=False):
        """Creates a Theano function that computes the output probabilities for
        a single time step.

        Creates the function self._step_function that takes as input a set of
        word sequences and the current recurrent states. It uses the previous
        states and word IDs to compute the output distributions, and computes
        the probabilities of the target words.

        All invocations of ``decode()`` will use the given NNLM weight and LM
        scale when computing the total probability. If LM scale is not given,
        uses the value provided in the lattice files. If it's not provided in a
        lattice file either, performs no scaling of LM log probabilities.

        ``decoding_options`` should countain the following elements:

        nnlm_weight : float
          weight of the neural network probabilities when interpolating with the
          lattice probabilities

        lm_scale : float
          if other than ``None``, the decoder will scale language model log
          probabilities by this factor; otherwise the scaling factor will be
          read from the lattice file

        wi_penalty : float
          penalize word insertion by adding this value to the total log
          probability of a token as many times as there are words

        unk_penalty : float
          if set to other than None, used as <unk> token score

        use_shortlist : bool
          if set to ``True``, <unk> token probability is distributed among the
          out-of-shortlist words according to their unigram probabilities

        unk_from_lattice : bool
          if set to ``True``, the probability for <unk> tokens is taken from the
          lattice alone

        linear_interpolation : bool
          if set to ``True``, use linear instead of (pseudo) log-linear
          interpolation of language model probabilities

        max_tokens_per_node : int
          if set to other than None, leave only this many tokens at each node

        beam : float
          if set to other than None, prune tokens whose total log probability is
          further than this from the best token at each point in time

        recombination_order : int
          number of words to consider when deciding whether two tokens should be
          recombined, or ``None`` for the entire word history

        prune_extra_limit : float
          if set, adjust the beam and max_tokens_per_node pruning relative to
          the number of tokens; the limits are divided by the number of tokens
          and multiplied by this factor

        abs_min_beam : float
          when using prune_extra_limit, this is the minimum that the beam will
          be adjusted to

        abs_min_max_tokens : float
          when using prune_extra_limit, this is the minimum that the maximum
          number of tokens will be adjusted to

        :type network: Network
        :param network: the neural network object

        :type decoding_options: dict
        :param decoding_options: a dictionary of decoding options (see above)

        :type profile: bool
        :param profile: if set to True, creates a Theano profile object
        """

        self._network = network
        self._vocabulary = network.vocabulary
        self._nnlm_weight = logprob_type(decoding_options['nnlm_weight'])
        self._lm_scale = decoding_options['lm_scale']
        self._wi_penalty = decoding_options['wi_penalty']
        self._unk_penalty = decoding_options['unk_penalty']
        self._unk_from_lattice = decoding_options['unk_from_lattice']
        self._linear_interpolation = decoding_options['linear_interpolation']
        self._max_tokens_per_node = decoding_options['max_tokens_per_node']
        self._beam = decoding_options['beam']
        if self._beam is not None:
            self._beam = logprob_type(self._beam)
        self._recombination_order = decoding_options['recombination_order']
        self._prune_extra_limit = decoding_options.get('prune_extra_limit', None)
        self._abs_min_beam = decoding_options.get('abs_min_beam', 0)
        self._abs_min_max_tokens = decoding_options.get('abs_min_max_tokens', 0)
        if self._prune_extra_limit is None:
            self._abs_min_beam = self._abs_min_max_tokens = 0

        if decoding_options['use_shortlist'] and \
           self._vocabulary.has_unigram_probs():
            oos_logprobs = numpy.log(self._vocabulary.get_oos_probs())
            self._oos_logprobs = oos_logprobs.astype(theano.config.floatX)
        else:
            self._oos_logprobs = None

        self._sos_id = self._vocabulary.word_to_id['<s>']
        self._eos_id = self._vocabulary.word_to_id['</s>']
        self._unk_id = self._vocabulary.word_to_id['<unk>']

        inputs = [network.input_word_ids,
                  network.input_class_ids,
                  network.target_class_ids]
        inputs.extend(network.recurrent_state_input)

        outputs = [tensor.log(network.target_probs())]
        outputs.extend(network.recurrent_state_output)

        # Ignore unused input, because is_training is only used by dropout
        # layer.
        self._step_function = theano.function(
            inputs,
            outputs,
            givens=[(network.is_training, numpy.int8(0))],
            name='step_predictor',
            profile=profile,
            on_unused_input='ignore')

    def decode(self, lattice):
        """Propagates tokens through given lattice and returns a list of tokens
        in the final nodes.

        Propagates tokens at a node to every outgoing link by creating a copy of
        each token and updating the language model scores according to the link.

        The function returns two lists. The first list contains the final
        tokens, sorted in the descending order of total log probability. I.e.
        the first token in the list represents the best path through the
        lattice. The second list contains the tokens that were dropped during
        recombination. This is needed for constructing a new rescored lattice.

        :type lattice: Lattice
        :param lattice: a word lattice to be decoded

        :rtype: a tuple of two lists of LatticeDecoder.Tokens
        :returns: a list of the final tokens sorted by probability (most likely
                  token first), and a list of the tokens that were dropped
                  during recombination
        """

        if self._lm_scale is not None:
            lm_scale = logprob_type(self._lm_scale)
        elif lattice.lm_scale is not None:
            lm_scale = logprob_type(lattice.lm_scale)
        else:
            lm_scale = logprob_type(1.0)

        if self._wi_penalty is not None:
            wi_penalty = logprob_type(self._wi_penalty)
        elif lattice.wi_penalty is not None:
            wi_penalty = logprob_type(lattice.wi_penalty)
        else:
            wi_penalty = logprob_type(0.0)

        tokens = [list() for _ in lattice.nodes]
        recomb_tokens = []
        initial_state = RecurrentState(self._network.recurrent_state_size)
        initial_token = self.Token(history=(self._sos_id,), state=initial_state)
        initial_token.recompute_hash(self._recombination_order)
        initial_token.recompute_total(self._nnlm_weight, lm_scale, wi_penalty,
                                      self._linear_interpolation)
        tokens[lattice.initial_node.id].append(initial_token)
        lattice.initial_node.best_logprob = initial_token.total_logprob

        sorted_nodes = lattice.sorted_nodes()
        self._nodes_processed = 0
        final_tokens = []
        for node in sorted_nodes:
            stats = self._prune(node, sorted_nodes, tokens, recomb_tokens)

            num_new_tokens = 0
            node_tokens = tokens[node.id]
            assert node_tokens
            if node.final:
                new_tokens = self._propagate(
                    node_tokens, None, lm_scale, wi_penalty)
                final_tokens.extend(new_tokens)
                num_new_tokens += len(new_tokens)
            for link in node.out_links:
                new_tokens = self._propagate(
                    node_tokens, link, lm_scale, wi_penalty)
                tokens[link.end_node.id].extend(new_tokens)
                # If there are lots of tokens in the end node, prune already to
                # conserve memory.
                if self._max_tokens_per_node is not None and \
                   len(tokens[link.end_node.id]) > self._max_tokens_per_node * 2:
                    self._prune(link.end_node, sorted_nodes, tokens, recomb_tokens)
                num_new_tokens += len(new_tokens)
            stats['new'] = num_new_tokens

            self._nodes_processed += 1
            self._log_stats(stats, node.id, len(sorted_nodes))

        if len(final_tokens) == 0:
            raise InputError("Could not reach a final node of word lattice.")

        final_tokens = self._sorted_recombined_tokens(final_tokens,
                                                      recomb_tokens)
        return final_tokens, recomb_tokens

    def _propagate(self, tokens, link, lm_scale, wi_penalty):
        """Propagates tokens to given link or to end of sentence.

        Lattices may contain null nodes with word ``None`` that model e.g.
        silence or sentence start or end, or for example when the topology is
        easier to represent with extra nodes. Such null nodes may contain
        language model scores. Then the function will update the acoustic and
        lattice LM score, but will not compute anything with the neural network.

        Also updates ``best_logprob`` of the end node, so that beam pruning
        threshold can be obtained efficiently.

        :type tokens: list of LatticeDecoder.Tokens
        :param tokens: input tokens

        :type link: Lattice.Link
        :param link: if other than ``None``, propagates the tokens to this link;
                     if ``None``, just updates the LM logprobs as if the tokens
                     were propagated to an end of sentence

        :type lm_scale: logprob_type
        :param lm_scale: scale language model log probabilities by this factor

        :type wi_penalty: logprob_type
        :param wi_penalty: penalize word insertion by adding this value to the
                           total log probability of the token

        :rtype: list of LatticeDecoder.Tokens
        :returns: the propagated tokens
        """

        new_tokens = [self.Token.copy(token) for token in tokens]

        if link is None:
            self._append_word(new_tokens, self._eos_id)
        else:
            for token in new_tokens:
                if link.ac_logprob is not None:
                    token.ac_logprob += link.ac_logprob
                if link.lm_logprob is not None:
                    token.lat_lm_logprob += link.lm_logprob

            if link.word is not None:
                try:
                    word = self._vocabulary.word_to_id[link.word]
                except KeyError:
                    word = link.word
                if self._unk_from_lattice:
                    self._append_word(new_tokens, word, link.lm_logprob)
                elif self._unk_penalty is not None:
                    self._append_word(new_tokens, word, self._unk_penalty)
                else:
                    self._append_word(new_tokens, word)

        for token in new_tokens:
            token.recompute_hash(self._recombination_order)
            token.recompute_total(self._nnlm_weight, lm_scale, wi_penalty,
                                  self._linear_interpolation)
            if link is not None and link.end_node is not None:
                if (link.end_node.best_logprob is None) or \
                   (token.total_logprob > link.end_node.best_logprob):
                    link.end_node.best_logprob = token.total_logprob

        return new_tokens

    def _prune(self, node, sorted_nodes, tokens, recomb_tokens):
        """Prunes tokens from a node according to beam and the maximum number of
        tokens, and recombines tokens whose N previous words are identical
        (where N is the recombination order).

        :type node: Lattice.Node
        :param node: perform pruning on this node

        :type sorted_nodes: list of Lattice.Nodes
        :param sorted_nodes: all nodes in topological order

        :type tokens: list of LatticeDecoder.Tokens
        :param tokens: all tokens

        :type recomb_tokens: list of tuples
        :param recomb_tokens: for all tokens that were dropped during
            recombination, contains the token, and the history and NNLM log
            probability of the kept token; dropped tokens will be added to the
            list

        :rtype: dict
        :returns: a dictionary of statistics collected during pruning
        """

        node_tokens = tokens[node.id]
        assert node_tokens
        stats = dict()
        stats['before'] = len(node_tokens)

        limit_divider = 1
        if self._prune_extra_limit is not None:
            limit_divider = len(sorted_nodes) // self._prune_extra_limit + 1
            limit_divider = max(1, limit_divider)

        new_tokens = self._sorted_recombined_tokens(node_tokens,
                                                    recomb_tokens)
        stats['after-recomb'] = len(new_tokens)

        # Compare to the best probability at the same or later time.
        if self._beam is not None:
            if node.time is None:
                node_ids = [iter_node.id for iter_node in sorted_nodes]
                time_begin = node_ids.index(node.id)
            else:
                for time_begin, iter_node in enumerate(sorted_nodes):
                    if (iter_node.time is not None) and \
                       (iter_node.time >= node.time):
                        break
            assert time_begin < len(sorted_nodes)
            best_logprob = max(iter_node.best_logprob
                               for iter_node in sorted_nodes[time_begin:]
                               if iter_node.best_logprob is not None)

            beam = self._beam / limit_divider
            beam = max(beam, self._abs_min_beam)
            threshold = best_logprob - beam

            stats['best'] = best_logprob
            stats['node-best'] = new_tokens[0].total_logprob
            stats['node-worst'] = new_tokens[-1].total_logprob
            stats['threshold'] = threshold
            stats['average'] = sum(t.total_logprob for t in new_tokens) / len(new_tokens)
            if len(new_tokens) >= 10:
                stats['pos10'] = new_tokens[9].total_logprob
            if len(new_tokens) >= 50:
                stats['pos50'] = new_tokens[49].total_logprob
            if len(new_tokens) >= 100:
                stats['pos100'] = new_tokens[99].total_logprob

            keep_tokens = len(new_tokens)
            while (keep_tokens > 1) and \
                  (new_tokens[keep_tokens - 1].total_logprob <= threshold):
                keep_tokens -= 1
            new_tokens = new_tokens[:keep_tokens]
        stats['after-beam'] = len(new_tokens)

        # Enforce limit on number of tokens at each node.
        if self._max_tokens_per_node is not None:
            max_tokens = self._max_tokens_per_node // limit_divider
            max_tokens = max(max_tokens, self._abs_min_max_tokens)
            new_tokens = new_tokens[:max_tokens]
        stats['after-max'] = len(new_tokens)

        tokens[node.id] = new_tokens
        return stats

    def _sorted_recombined_tokens(self, tokens, recomb_tokens):
        """Sorts tokens by descending probability and recombines tokens with
        identical hash.

        First sorts the tokens by probability, most likely token first. Then
        keeps only the first if there are multiple tokens with the same hash.
        Tokens with identical N previous words in the history (where N is the
        recombination order) will be dropped and added to recomb_tokens.

        :type tokens: list of LatticeDecoder.Tokens
        :param tokens: input tokens

        :type recomb_tokens: list of tuples
        :param recomb_tokens: for all tokens that were dropped during
            recombination, contains the token, and the history and NNLM log
            probability of the kept token; dropped tokens will be added to the
            list

        :rtype: list of LatticeDecoder.Tokens
        :returns: the tokens in sorted order and with only the best token of
                  those with identical recombination hash
        """

        sorted_tokens = sorted(tokens,
                               key=lambda token: token.total_logprob,
                               reverse=True)

        token_map = dict()
        result = []
        for token in sorted_tokens:
            key = token.recombination_hash
            if key not in token_map:
                result.append(token)
                token_map[key] = token
            else:
                kept_token = token_map[key]
                recomb_tokens.append((token,
                                      kept_token.history,
                                      kept_token.nn_lm_logprob))
        return result

    def _append_word(self, tokens, target_word, oov_logprob=None):
        """Appends a word to each of the given tokens, and updates their scores.

        :type tokens: list of LatticeDecoder.Tokens
        :param tokens: input tokens

        :type target_word: int or str
        :param target_word: word ID or word to be appended to the existing
                            history of each input token; if not an integer, the
                            word will be considered ``<unk>`` and this variable
                            will be taken literally as the word that will be
                            used in the resulting transcript

        :type oov_logprob: float
        :param oov_logprob: log probability to be assigned to OOV words
        """

        def limit_to_shortlist(self, word):
            """Returns the ``<unk>`` word ID if the argument is not a shortlist
            word ID.
            """
            if isinstance(word, int) and self._vocabulary.in_shortlist(word):
                return word
            else:
                return self._unk_id

        input_word_ids = [[limit_to_shortlist(self, token.history[-1])
                           for token in tokens]]
        input_word_ids = numpy.asarray(input_word_ids).astype('int64')
        input_class_ids, membership_probs = \
            self._vocabulary.get_class_memberships(input_word_ids)
        recurrent_state = [token.state for token in tokens]
        recurrent_state = RecurrentState.combine_sequences(recurrent_state)
        target_word_id = limit_to_shortlist(self, target_word)
        target_class_ids = numpy.ones(shape=(1, len(tokens))).astype('int64')
        target_class_ids *= self._vocabulary.word_id_to_class_id[target_word_id]
        step_result = self._step_function(input_word_ids,
                                          input_class_ids,
                                          target_class_ids,
                                          *recurrent_state.get())
        logprobs = step_result[0]
        # Add logprobs from the class membership of the predicted words.
        logprobs += numpy.log(membership_probs)
        output_state = step_result[1:]

        for index, token in enumerate(tokens):
            token.history = token.history + (target_word,)
            token.state = RecurrentState(self._network.recurrent_state_size)
            # Slice the sequence that corresponds to this token.
            token.state.set([layer_state[:, index:index + 1]
                             for layer_state in output_state])
            # logprobs matrix contains only one time step.
            token.nn_lm_logprob += self._handle_unk_logprob(target_word,
                                                            logprobs[0, index],
                                                            oov_logprob)

    def _handle_unk_logprob(self, word, network_logprob, oov_logprob):
        """Returns the log probability after applying <unk> processing.

        If ``self._oos_logprobs`` is set and the word is in vocabulary, the
        corresponding value will be added to the network log probability. In
        effect, the probability of out-of-shortlist words (which is the <unk>
        probability) is multiplied by the fraction of the actual word within the
        set of OOS words. For out-of-vocabulary words returns ``oov_logprob`` or
        the value predicted by the network.

        Otherwise, for both out-of-shortlist and out-of-vocabulary words returns
        ``oov_logprob`` or the value predicted by the network.

        For shortlist words returns the value predicted by the network.

        :type word: int or str
        :param word: target word ID or word; if not an integer, the word will be
                     considered ``<unk>``

        :type network_logprob: float
        :param network_logprob: log probability predicted by the network

        :type oov_logprob: float
        :param oov_logprob: log probability to be assigned to OOV words
        """

        in_vocabulary = isinstance(word, int)
        in_shortlist = in_vocabulary and self._vocabulary.in_shortlist(word)

        if self._oos_logprobs is not None:
            if in_vocabulary:
                return network_logprob + self._oos_logprobs[word]
            elif oov_logprob is not None:
                logging.debug("Replacing <unk> logprob %f with %f.",
                              network_logprob, oov_logprob)
                return oov_logprob
        elif (not in_shortlist) and (oov_logprob is not None):
            logging.debug("Replacing <unk> logprob %f with %f.",
                          network_logprob, oov_logprob)
            return oov_logprob

        return network_logprob

    def _log_stats(self, stats, node_id, num_nodes):
        """Writes pruning statistics to debug log.

        :type stats: dict
        :param stats: a dictionary of statistics on the number of tokens and
                      their log probabilities
        """

        if self._nodes_processed % math.ceil(num_nodes / 20) != 0:
            return

        optional = ''
        for name in ('after-recomb', 'after-beam', 'after-max', 'new'):
            if name in stats:
                optional += ' {}={}'.format(name, stats[name])
        logging.debug('[%d] (%.2f %%) node=%d -- tokens before=%d%s',
                      self._nodes_processed,
                      self._nodes_processed / num_nodes * 100,
                      node_id,
                      stats['before'],
                      optional)

        if 'best' in stats:
            optional = ''
            for name in ('node-best', 'node-worst', 'threshold', 'average',
                         'pos10', 'pos50', 'pos100'):
                if name in stats:
                    optional += ' {}={:.1f}'.format(name, stats[name])
            logging.debug('logprob best=%.1f%s',
                          stats['best'],
                          optional)



"""Modules implementing the theanolm subcommands.
"""

import theanolm.commands.train
import theanolm.commands.score
import theanolm.commands.decode
import theanolm.commands.sample
import theanolm.commands.version



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that extends the operations provided by Theano.

The functions are designed to be compatible with the Keras interface.
"""

import theano.tensor as tensor
from theano.gof import MissingInputError

def conv2d(input_matrix, filters, strides=(1, 1), padding='valid'):
    """Performs 2-dimensional convolution given 4-dimensional input and filter
    tensors.

    The input is a mini-batch of 2-dimensional n-channel samples. Convolution is
    applied independently to each sample.

    The operation can add padding to the input tensor in order apply the filter
    also in places where it does not completely overlap with the input. At least
    the following modes are supported:

      - ``same``: Apply so much that the output dimensions will be the same as
                  the input dimensions.
      - ``valid``: Only apply the filter where all elements overlap with the
                   input. The output will be smaller than the input.
      - ``full``: Apply the filter where any part of it overlaps with the input.
                  The output will be larger than the input.

    :type input_matrix: symbolic 4D tensor
    :param input_matrix: a mini-batch of shape (samples, rows, columns, features)

    :type filters: symbolic 3D tensor
    :param filters: filters of shape (rows, columns, input features, output
                    features)

    :type strides: int
    :param strides: the steps in which the sliding window is shifted in each
                    dimension

    :type padding: str
    :param padding: padding to be applied to the input
    """

    # Permute input dimensions from (samples, rows, columns, features) to
    # (samples, features, rows, columns).
    input_matrix = input_matrix.dimshuffle(0, 3, 1, 2)
    # Permute filter dimensions from (rows, columns, input features, output
    # features) to (output features, input features, rows, columns)`.
    filters = filters.dimshuffle(3, 2, 0, 1)
    # Convert padding to border_mode values understood by Theano.
    if padding == 'same':
        border_mode = 'half'
    elif padding == 'valid':
        border_mode = 'valid'
    elif padding == 'full':
        border_mode = 'full'
    else:
        raise ValueError('Unknown input padding requested: {}'.format(padding))
    # If filters is a shared variable, we can get its shape and use it to select
    # an optimal implementation.
    try:
        filter_shape = filters.eval().shape
    except MissingInputError:
        filter_shape = None

    result = tensor.nnet.conv2d(input_matrix,
                                filters,
                                border_mode=border_mode,
                                subsample=strides,
                                input_shape=None,
                                filter_shape=filter_shape,
                                filter_flip=False)

    if padding == 'same':
        # If the filter width or height is not an odd number, there will be an
        # extra element in the output.
        rows_end = (input_matrix.shape[2] + strides[0] - 1) // strides[0]
        result = result[:, :, :rows_end, :]
        columns_end = (input_matrix.shape[3] + strides[1] - 1) // strides[1]
        result = result[:, :, :, :columns_end]

    # Permute output dimensions from (samples, features, rows, columns) to
    # (samples, rows, columns, features).
    return result.dimshuffle(0, 2, 3, 1)

def conv1d(input_matrix, filters, strides=1, padding='valid'):
    """Performs 1-dimensional convolution given 3-dimensional input and filter
    tensors.

    The input is a mini-batch of 1-dimensional n-channel samples. Convolution is
    applied independently to each sample.

    Adds a dimension of size 1 to input and filter tensors and calls `conv2d`.
    Compatible with the Keras interface.

    The operation can add padding to the input tensor in order apply the filters
    also in places where it does not completely overlap with the input. At least
    the following modes are supported:

      - ``same``: Apply so much that the output dimensions will be the same as
                  the input dimensions.
      - ``valid``: Only apply the filters where all elements overlap with the
                   input. The output will be smaller than the input.
      - ``full``: Apply the filters where any part of it overlaps with the input.
                  The output will be larger than the input.

    :type input_matrix: symbolic 3D tensor
    :param input_matrix: a mini-batch of shape (samples, elements, features)

    :type filters: symbolic 3D tensor
    :param filters: filters of shape (elements, input features, output features)

    :type strides: int
    :param strides: the steps in which the sliding window is shifted

    :type padding: str
    :param padding: padding to be applied to the input
    """

    # Permute input dimensions from (samples, elements, features) to
    # (samples, rows, 1, features).
    input_matrix = input_matrix.dimshuffle(0, 1, 'x', 2)
    # Permute filter dimensions from (elements, input features, output features)
    # to (rows, 1, input features, output features).
    filters = filters.dimshuffle(0, 'x', 1, 2)
    # Stride dimensions from (strides) to (strides, 1).
    strides = (strides, 1)

    output = conv2d(input_matrix, filters, strides=strides, padding=padding)

    # Output dimensions from (samples, rows, 1, features) to
    # (samples, elements, features).
    return output.reshape([output.shape[0], output.shape[1], output.shape[3]])

def l1_norm(tensors):
    """Returns the sum of absolute values.

    :type tensors: iterable of symbolic tensors
    :param tensors: the function will compute the sum over all the elements of
                    these tensors

    :rtype: symbolic float
    :return: L1 norm of all the elements in all the tensors
    """

    absolutes = [abs(x) for x in tensors]
    return sum(tensor.sum(x) for x in absolutes)

def sum_of_squares(tensors):
    """Returns the sum of squared values.

    :type tensors: iterable of symbolic tensors
    :param tensors: the function will compute the sum over all the elements of
                    these tensors

    :rtype: symbolic float
    :return: sum of squares of all the elements in all the tensors
    """

    squares = [tensor.sqr(x) for x in tensors]
    return sum(tensor.sum(x) for x in squares)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the "theanolm version" command.
"""

import theano

from theanolm import __version__

def version(args):
    """A function that performs the "theanolm version" command.

    :type args: argparse.Namespace
    :param args: a collection of command line arguments
    """

    print("TheanoLM", __version__)
    print("Theano", theano.version.version)
    try:
        import pygpu
        print("pygpu", pygpu.__version__)
    except ImportError:
        print("No pygpu")




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the Lattice class used by word lattice decoders.
"""

import logging

from theanolm.backend import InputError

class NodeNotFoundError(Exception):
    pass

class Lattice(object):
    """Word Lattice

    Word lattice describes a search space for decoding. The graph is represented
    as a list of nodes and links. Each node contains pointers to its incoming
    and outgoing links. Each link contains a pointer to the nodes in both ends.
    """

    class Link(object):
        """A link between two graph nodes.

        A link contains pointers to the start and end node. A node that has the
        link as an outgoing link can find the next node from ``end_node`` and a
        node that has the link as an incoming link can find the previous node
        from ``start_node``.
        """

        def __init__(self, start_node, end_node, word=None,
                     ac_logprob=None, lm_logprob=None, transitions=""):
            """Constructs a link.

            :type start_node: self.Node
            :param start_node: the node that has this link as an outgoing link

            :type end_node: self.Node
            :param end_node: the node that has this link as an incoming link

            :type word: str or int
            :param word: the word label on the link

            :type ac_logprob: float
            :param ac_logprob: acoustic log probability

            :type lm_logprob: float
            :param lm_logprob: language model log probability

            :type transitions: str
            :param transitions: transitions for an FST lattice
            """

            self.start_node = start_node
            self.end_node = end_node
            self.word = word
            self.ac_logprob = ac_logprob
            self.lm_logprob = lm_logprob
            self.transitions = transitions

    class Node(object):
        """A node in the graph.

        Outgoing and incoming links can be used to find the next and previous
        nodes in the topology.
        """

        def __init__(self, node_id):
            """Constructs a node with no links.

            :type node_id: int
            :param node_id: the ID that can be used to access the node in the
                            node list
            """

            self.id = node_id
            self.out_links = []
            self.in_links = []
            self.time = None
            self.best_logprob = None
            self.final = False

    def __init__(self):
        """Constructs an empty lattice.
        """

        self.nodes = []
        self.links = []
        self.initial_node = None
        self.utterance_id = None
        self.lm_scale = None
        self.wi_penalty = None

    def write_slf(self, output_file):
        """Writes the lattice in SLF format.

        :type output_file: file object
        :param output_file: a file where to write the output
        """

        output_file.write("# Header (generated by TheanoLM)\n")
        output_file.write("VERSION=1.1\n")
        output_file.write('UTTERANCE="{}"\n'.format(self.utterance_id))
        fields = []
        if self.lm_scale is not None:
            fields.append("lmscale={}".format(self.lm_scale))
        if self.wi_penalty is not None:
            fields.append("wdpenalty={}".format(self.wi_penalty))
        if fields:
            output_file.write("\t".join(fields) + "\n")
        output_file.write("start={}\n".format(self.initial_node.id))
        output_file.write("N={}\tL={}\n"
                          .format(len(self.nodes), len(self.links)))

        output_file.write("# Nodes\n")
        for node in self.nodes:
            fields = ["I={}".format(node.id)]
            if node.time is not None:
                fields.append("t={}".format(node.time))
            output_file.write("\t".join(fields) + "\n")

        output_file.write("# Links\n")
        for link_id, link in enumerate(self.links):
            fields = ["J={}".format(link_id),
                      "S={}".format(link.start_node.id),
                      "E={}".format(link.end_node.id)]
            if link.word is None:
                fields.append("W=!NULL")
            else:
                fields.append("W={}".format(link.word))
            if link.ac_logprob is not None:
                fields.append("a={}".format(link.ac_logprob + 0.0))
            if link.lm_logprob is not None:
                fields.append("l={}".format(link.lm_logprob + 0.0))
            output_file.write("\t".join(fields) + "\n")

    def write_kaldi(self, output_file, word_to_id):
        """Writes the lattice in Kaldi CompactLattice format.

        :type output_file: file object
        :param output_file: a file where to write the output

        :type word_to_id: Vocabulary
        :param word_to_id: mapping of words to Kaldi IDs
        """

        def write_normal_link(link):
            word = link.word
            if word is None:
                word = "<eps>"
            output_file.write("{} {} {} {},{},{}\n".format(
                link.start_node.id,
                link.end_node.id,
                word_to_id[word],
                -link.lm_logprob + 0.0,
                -link.ac_logprob + 0.0,
                link.transitions))

        def write_final_link(link):
            output_file.write("{} {},{},{}\n".format(
                link.start_node.id,
                -link.lm_logprob + 0.0,
                -link.ac_logprob + 0.0,
                link.transitions))

        output_file.write("{}\n".format(self.utterance_id))
        for node in self.nodes:
            for link in node.out_links:
                if link.end_node.final:
                    write_final_link(link)
                else:
                    write_normal_link(link)
        output_file.write("\n")

    def sorted_nodes(self):
        """Sorts nodes topologically, then by time.

        Returns a list which contains the nodes in sorted order. Uses the Kahn's
        algorithm to sort the nodes topologically, but always picks the node
        from the queue that has the lowest time stamp, if the nodes contain time
        stamps.
        """

        result = []
        # A queue of nodes to be visited next:
        node_queue = [self.initial_node]
        # The number of incoming links not traversed yet:
        in_degrees = [len(node.in_links) for node in self.nodes]
        while node_queue:
            node = node_queue.pop()
            result.append(node)
            for link in node.out_links:
                next_node = link.end_node
                in_degrees[next_node.id] -= 1
                if in_degrees[next_node.id] == 0:
                    node_queue.append(next_node)
                    node_queue.sort(key=lambda x: (x.time is None, x.time),
                                    reverse=True)
                elif in_degrees[next_node.id] < 0:
                    raise InputError("Word lattice contains a cycle.")

        if len(result) < len(self.nodes):
            logging.warning("Word lattice contains unreachable nodes.")
        else:
            assert len(result) == len(self.nodes)

        return result

    def _add_link(self, start_node, end_node):
        """Adds a link between two nodes.

        :type start_node: Node
        :param start_node: creates a link from this node

        :type end_node: Node
        :param end_node: creates a link to this node

        :rtype: Link
        :returns: the created link
        """

        link = self.Link(start_node, end_node)
        self.links.append(link)
        start_node.out_links.append(link)
        end_node.in_links.append(link)
        return link



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Functions related to reading text.
"""

def utterance_from_line(line):
    """Converts a line of text, read from an input file, into a list of words.

    Start-of-sentence and end-of-sentece tokens (``<s>`` and ``</s>``) will be
    inserted at the beginning and the end of the list, if they're missing. If
    the line is empty, returns an empty list (instead of an empty sentence
    ``['<s>', '</s>']``).

    :type line: str or bytes
    :param line: a line of text (read from an input file)

    :rtype: list of strs
    :returns: list of words / tokens
    """

    if isinstance(line, bytes):
        line = line.decode('utf-8')
    line = line.rstrip()
    if not line:
        # empty line
        return []

    result = line.split()
    if result[0] != '<s>':
        result.insert(0, '<s>')
    if result[-1] != '</s>':
        result.append('</s>')

    return result

def find_sentence_starts(data):
    """Finds the positions inside a memory-mapped file, where the sentences
    (lines) start.

    TextIOWrapper disables tell() when readline() is called, so search for
    sentence starts in memory-mapped data.

    :type data: mmap.mmap
    :param data: memory-mapped data of the input file

    :rtype: list of ints
    :returns: a list of file offsets pointing to the next character from a
              newline (including file start and excluding file end)
    """

    result = [0]

    pos = 0
    while True:
        pos = data.find(b'\n', pos)
        if pos == -1:
            break
        pos += 1
        if pos < len(data):
            result.append(pos)

    return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the BasicOptimizer class, a base class for
optimizers.
"""

from abc import abstractmethod, ABCMeta

import numpy
import theano
import theano.tensor as tensor

from theanolm.backend import IncompatibleStateError
from theanolm.backend import test_value
from theanolm.backend import sum_of_squares

class BasicOptimizer(object, metaclass=ABCMeta):
    """Superclass for Neural Network Language Model Optimizers
    """

    def __init__(self, optimization_options, network, cost_function,
                 profile=False):
        """Creates Theano functions for training a neural network language
        model.

        The subclass constructor is expected to create the optimizer parameters
        in ``self._params``. This constructor will then create a function
        ``self.update_function``, which updates the optimizer parameters, and
        then the model state given the gradients, the optimizer parameters, and the
        learning rate.

        The update functions takes as arguments four matrices and the alpha
        hyperparameter:

        1. Word IDs in the shape of a mini-batch. The functions will slice this
           into input and output.
        2. Class IDs in the shape of a mini-batch. The functions will slice this
           into input and output.
        3. Mask in the shape of a mini-batch, but only for the output words (not
           for the first time step).
        4. Weights in the shape of a mini-batch, but only for the output words
           (not for the first time step).
        4. Alpha or learning rate is used to scale the size of the update.

        :type optimization_options: dict
        :param optimization_options: a dictionary of optimization options

        :type network: Network
        :param network: the neural network object

        :type cost_function: Cost
        :param cost_function: an object from one of the cost function classes
                              that defined the training objective

        :type profile: bool
        :param profile: if set to True, creates a Theano profile object
        """

        self.network = network

        float_type = numpy.dtype(theano.config.floatX).type
        self.float_type = float_type

        try:
            # numerical stability / smoothing term to prevent divide-by-zero
            self._epsilon = float_type(optimization_options['epsilon'])
            # learning rate / step size
            self.learning_rate = float_type(optimization_options['learning_rate'])
            # weights for training files
            self._weights = optimization_options['weights']
            # maximum norm for parameter updates
            self._max_gradient_norm = float_type(
                optimization_options['max_gradient_norm'])
            # number of noise samples for sampling based output
            num_noise_samples = optimization_options['num_noise_samples']
            # noise sample sharing for sampling based output
            noise_sharing = optimization_options['noise_sharing']
        except KeyError as e:
            raise ValueError("Option {} is missing from optimization options."
                             .format(e))

        self._unk_id = self.network.vocabulary.word_to_id['<unk>']

        # The function takes as inputs a mini-batch of word IDs and class IDs,
        # and slices the input and target IDs for the network.
        batch_word_ids = tensor.matrix('optimizer/batch_word_ids',
                                       dtype='int64')
        batch_word_ids.tag.test_value = test_value(
            size=(101, 16), high=self.network.vocabulary.num_shortlist_words())
        batch_class_ids = tensor.matrix('optimizer/batch_class_ids',
                                        dtype='int64')
        batch_class_ids.tag.test_value = test_value(
            size=(101, 16), high=self.network.vocabulary.num_classes())

        # Derive the symbolic expression for updating the gradient with regard
        # to each parameter.
        cost, num_words = cost_function.get_tensor()
        self._gradients = \
            tensor.grad(cost, wrt=list(self.network.get_variables().values()))

        # The function takes as input the learning rate.
        alpha = tensor.scalar('optimizer/alpha',
                              dtype=theano.config.floatX)
        alpha.tag.test_value = 0.1

        # The function takes as input a matrix of weights, one for each
        # target word. These are used to scale the parameter updates.
        weights = tensor.matrix('optimizer/weights',
                                dtype=theano.config.floatX)
        weights.tag.test_value = test_value(size=(100, 16), high=1.0)
        word_positions = tensor.eq(self.network.mask, 1).nonzero()
        weight = weights[word_positions].sum()
        num_words_float = tensor.cast(num_words, theano.config.floatX)
        modified_alpha = tensor.switch(tensor.gt(num_words, 0),
                                       alpha * weight / num_words_float,
                                       alpha)

        # Ignore unused input, because is_training is only used by dropout
        # layer.
        self.update_function = theano.function(
            [batch_word_ids, batch_class_ids, self.network.mask, weights,
             alpha],
            [cost, num_words],
            givens=[(network.input_word_ids, batch_word_ids[:-1]),
                    (network.input_class_ids, batch_class_ids[:-1]),
                    (network.target_word_ids, batch_word_ids[1:]),
                    (network.target_class_ids, batch_class_ids[1:]),
                    (self.network.is_training, numpy.int8(1)),
                    (self.network.num_noise_samples,
                     numpy.int64(num_noise_samples))],
            updates=self._get_param_updates(alpha),
            name='update_function',
            on_unused_input='ignore',
            profile=profile)

    def get_state(self, state):
        """Pulls parameter values from Theano shared variables.

        If there already is a parameter in the state, it will be replaced, so it
        has to have the same number of elements.

        :type state: h5py.File
        :param state: HDF5 file for storing the optimization parameters
        """

        h5_optimizer = state.require_group('optimizer')
        h5_optimizer.attrs['learning_rate'] = self.learning_rate

        self._params.get_state(state)

    def set_state(self, state):
        """Sets the values of Theano shared variables.

        Requires that ``state`` contains values for all the optimization
        parameters.

        :type state: h5py.File
        :param state: HDF5 file that contains the optimization parameters
        """

        if 'optimizer' not in state:
            raise IncompatibleStateError("Optimizer state is missing.")
        h5_optimizer = state['optimizer']

        if 'learning_rate' not in h5_optimizer.attrs:
            raise IncompatibleStateError("Learning rate is missing from "
                                         "optimizer state.")
        self.learning_rate = h5_optimizer.attrs['learning_rate']

        self._params.set_state(state)

    def update_minibatch(self, word_ids, class_ids, file_ids, mask):
        """Optimizes the neural network parameters using the given inputs and
        learning rate.

        :type word_ids: ndarray of ints
        :param word_ids: a 2-dimensional matrix, indexed by time step and
                         sequence, that contains the word IDs

        :type class_ids: ndarray of ints
        :param class_ids: a 2-dimensional matrix, indexed by time step and
                          sequence, that contains the class IDs

        :type file_ids: ndarray of ints
        :param file_ids: a 2-dimensional matrix, indexed by time step and
                         sequence, that identifies the file in case of multiple
                         training files

        :type mask: numpy.ndarray of a floating point type
        :param mask: a 2-dimensional matrix, indexed by time step and sequence,
                     that masks out elements past the sequence ends.
        """

        # We should predict probabilities of the words at the following time
        # step.
        mask = mask[1:]
        file_ids = file_ids[1:]
        weights = self._weights[file_ids]
        alpha = self.learning_rate
        self.update_function(word_ids, class_ids, mask, weights, alpha)

    @abstractmethod
    def _get_param_updates(self, alpha):
        """Returns Theano expressions for updating the model parameters and any
        additional parameters required by the optimizer. Implemented by every
        optimizer subclass.

        :type alpha: Variable
        :param alpha: a scale to be applied to the model parameter updates

        :rtype: iterable over pairs (shared variable, new expression)
        :returns: expressions how to update the optimizer parameters
        """

        assert False

    def _normalize(self, deltas):
        """Normalizes the norm of parameter updates to given maximum value.

        :type deltas: dict of strs to symbolic tensors
        :param deltas: mapping from variable names to symbolic tensors that
                       describe the amount and direction of parameter updates
                       (normally the negative gradient of each parameter), after
                       any adaptation applied by the optimization method

        :rtype: dict of strs to symbolic tensors
        :returns: mapping from variable names to symbolic tensors that describe
                  ``deltas`` after normalization has been applied
        """

        max_norm = self._max_gradient_norm
        if max_norm is not None:
            norm = tensor.sqrt(sum_of_squares(deltas.values()))
            target_norm = tensor.clip(norm, 0.0, max_norm)
            for name, delta in deltas.items():
                deltas[name] = delta * target_norm / (self._epsilon + norm)
        return deltas



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the RecurrentState class, for storing the hidden
state of recurrent layers.
"""

import numpy
import theano

class RecurrentState:
    """State of Recurrent Layers at Certain Time Step

    Recurrent layers contain one or more state variables (vectors). When
    performing a forward pass one time step at a time, these state variables
    have to be saved and passed as input to the next time step. A
    ``RecurrentState`` object contains a matrix for each recurrent
    state variable.

    Every matrix has the same shape as a mini-batch: First dimension is the time
    step. Since one time step is processed at a time, the size of the first
    dimension is one. The second dimension is sequence. Multiple sequences may
    be processed in parallel. The third dimension is the state vector.
    """

    def __init__(self, sizes, num_sequences=1, state_variables=None):
        """Constructs a list of recurrent layer states and initializes them.

        This will create state vectors for each layer for each of the sequences
        that will be processed in parallel. Unless ``state_variables`` is given,
        the vectors will be initialized to zeros.

        :type sizes: list of ints
        :param sizes: size of each recurrent layer state

        :type num_sequences: int
        :param num_sequences: number of sequences to be processed in parallel

        :type state_variables: list of numpy.ndarrays
        :param state_variables: if set to other than ``None``, sets the initial
                                recurrent layer states to this instead of zeros
        """

        self.sizes = sizes
        self.num_sequences = num_sequences
        if state_variables is None:
            self.reset()
        else:
            self.set(state_variables)

    @classmethod
    def combine_sequences(cls, states):
        """Creates recurrent state variables that combine all the sequences from
        the given list of state objects.

        Takes multiple ``RecurrentState`` objects and creates one that contains
        a state matrix or matrices that combine all the sequences. The resulting
        state will contain as many sequences as the input states in total, in
        the same order as in the ``states`` list. The purpose is to be able to
        process multiple forward passes in parallel.

        :type states: list of RecurrentStates
        :param states: list of recurrent layer states, each containing N1, N2,
                       N3, ... sequences

        :rtype: RecurrentState
        :returns: a recurrent layer state that contains the N1 + N2 + N3 + ...
                  sequences from ``states``, in the same order that they appear
                  in ``states``
        """

        if not states:
            raise ValueError("Need at least one RecurrentState object.")

        num_sequences = sum([state.num_sequences for state in states])

        sizes = []
        state_variables = []
        num_variables = len(states[0].sizes)
        for index in range(num_variables):
            state_variable = [state.get(index) for state in states]
            state_variable = numpy.concatenate(state_variable, axis=1)
            assert state_variable.shape[0] == 1
            assert state_variable.shape[1] == num_sequences
            assert state_variable.shape[2] == states[0].sizes[index]
            sizes.append(state_variable.shape[2])
            state_variables.append(state_variable)

        return cls(sizes, num_sequences, state_variables)

    def reset(self):
        """Resets the state of each recurrent layer to zeros.

        The state vectors are 3-dimensional, because the layers support
        mini-batches with multiple sequences and time steps. Currently only one
        sequence can be computed at a time, when passing a state from time step
        to the next. Thus the first two dimensions have size 1.
        """

        self._state_variables = []
        for size in self.sizes:
            shape = (1, self.num_sequences, size)
            value = numpy.zeros(shape).astype(theano.config.floatX)
            self._state_variables.append(value)

    def set(self, state_variables):
        """Sets the state vector of every recurrent layer.

        :type state_variables: list of numpy.ndarrays
        :param state_variables: a matrix for each recurrent layer that contains
                                the state vector for each sequence at one time
                                step
        """

        if len(state_variables) != len(self.sizes):
            raise ValueError("Recurrent state should contain as many arrays "
                             "as there are recurrent layers.")
        for state_variable, size in zip(state_variables, self.sizes):
            if state_variable.shape[0] != 1:
                raise ValueError("Recurrent state should contain only one time "
                                 "step.")
            if state_variable.shape[1] != self.num_sequences:
                raise ValueError("Recurrent state contains incorrect number of "
                                 "sequences.")
            if state_variable.shape[2] != size:
                raise ValueError("Recurrent state contains a layer with "
                                 "incorrect size.")
        self._state_variables = state_variables

    def get(self, index=None):
        """Returns the state matrix of a given layer, or a list of the matrices
        of all layers.

        :type index: int
        :param index: index of a recurrent layer in the state object; if
                            set to other than ``None``, returns only the matrix
                            for the corresponding layer

        :rtype: numpy.ndarray or list of numpy.ndarrays
        :returns: a matrix for each recurrent layer that contains the state
                  vector for each sequence at one time step
        """

        if index is not None:
            return self._state_variables[index]
        else:
            return self._state_variables



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the Nesterov momentum optimizer.
"""

import numpy

from theanolm.backend import Parameters
from theanolm.training.basicoptimizer import BasicOptimizer

class NesterovOptimizer(BasicOptimizer):
    """Nesterov Momentum Optimization Method

    Normally Nesterov momentum is implemented by first taking a step towards
    the previous update direction, calculating gradient at that position,
    using the gradient to obtain the new update direction, and finally
    updating the parameters. We use an alternative formulation that requires
    the gradient to be computed only at the current parameter values,
    described here:
    https://github.com/lisa-lab/pylearn2/pull/136#issuecomment-10381617

    v_{t} = mu * v_{t-1} - lr * gradient(params_{t-1})
    params_{t} = params_{t-1} + mu * v_{t} - lr * gradient(params_{t-1})
    """

    def __init__(self, optimization_options, network, *args, **kwargs):
        """Creates a Nesterov momentum optimizer. Nesterov momentum optimizer
        does not use additional parameters.

        :type optimization_options: dict
        :param optimization_options: a dictionary of optimization options

        :type network: Network
        :param network: the neural network object
        """

        self._params = Parameters()
        for path, param in network.get_variables().items():
            self._params.add(path + '_velocity',
                             numpy.zeros_like(param.get_value()))

        # momentum
        if 'momentum' not in optimization_options:
            raise ValueError("Momentum is not given in optimization options.")
        self._momentum = optimization_options['momentum']

        super().__init__(optimization_options, network, *args, **kwargs)

    def _get_param_updates(self, alpha):
        """Returns Theano expressions for updating the model parameters and any
        additional parameters required by the optimizer.

        :type alpha: Variable
        :param alpha: a scale to be applied to the model parameter updates

        :rtype: iterable over pairs (shared variable, new expression)
        :returns: expressions how to update the optimizer parameters
        """

        result = []
        deltas = dict()
        for path, gradient in zip(self.network.get_variables(),
                                  self._gradients):
            deltas[path] = -gradient
        self._normalize(deltas)

        for path, param_old in self.network.get_variables().items():
            delta = deltas[path]
            velocity_old = self._params[path + '_velocity']
            velocity = self._momentum * velocity_old + alpha * delta
            param = param_old + self._momentum * velocity + alpha * delta
            result.append((velocity_old, velocity))
            result.append((param_old, param))
        return result
